import lhctodd as m


def test_version():
    assert m.__version__

def test_data():
    try:
        m.list()
    except:
        raise RuntimeError('Failed to fetch data from database')

def test_thoery():
    # check alpha_s
    print("as(MZ) = ", m.theory.width._as(9.1200000e+01))
    print("Gamma_vector_qq = ", m.theory.width.vector_qq(100))





import numpy as np
import os
from pathlib import Path

class dd_format:
    def __init__(self, limit, meta):
        self.ncol = limit.shape[1]
        self.size = limit.shape[0]

        self.limit = limit.tobytes()
        self.cite = meta.get("cite", "None")
        self.year = meta.get("year", "None")
        self.type = meta.get("type", "None")
        self.expr = meta.get("expr", "None")
        self.name = meta.get("name", "None")

    def get_limit(self):
        limit = np.frombuffer(self.limit, dtype=np.float64)
        return limit.reshape(self.size, self.ncol)

    def __str__(self):
        return "https://arxiv.org/abs/{0} {1} {2}".format(self.cite, self.type, self.name)


__data_path__ = Path(os.path.join(os.path.dirname(__file__), "data"))



from .tools import __data_path__
from scipy import interpolate
import numpy as np

class width:
    # quark masses from PDG 2020
    _q_mass = [
        0.00216, # up
        0.00467, # down
        0.093,   # strange
        1.27,    # charm
        4.18,    # bottom
        172.76   # top
    ]
    # lepton masses: e mu tau from PDG 2020
    _l_mass = [0.000511, 0.105658, 1.77682]
    _vev = 246
    
    # fermion-Yukawa-Coupling function 
    _fyc = lambda mf: np.sqrt(2)*mf/_vev

    # fetch alpha_s values taken from NNPDF31
    _as_data = np.genfromtxt(
        str( __data_path__ / f"alpha_s.csv"),
        delimiter=","
    )
    _as = interpolate.interp1d(
        _as_data[:,0], # scale
        _as_data[:,1], # alpha_s value
        fill_value="extrapolate"
    )

    # Vector mediators
    @classmethod
    def vector_qq(cls, med_mass, g=1.0):
        """ Width of vector mediator decaying to quarks        
        """
        width = 0
        for mass in cls._q_mass:
            z = np.divide(mass, med_mass)**2
            z = z.astype(np.complex128)
            width += np.where(
                med_mass >= 2 * mass,
                g**2 * med_mass * np.sqrt(1 - 4*z) * (1 + 2*z) / (4*np.pi),
                0.0 
            )
        return width

    @classmethod
    def vector_ll(cls, med_mass, g=1.0):
        """Width of vector mediator decaying to leptons
        """
        width = 0
        for mass in cls._l_mass:
            z = np.divide(mass, med_mass)**2
            z = z.astype(np.complex128)
            width += np.where(
                med_mass >= 2 * mass,
                g**2 * med_mass * np.sqrt(1 - 4*z) * (1 + 2*z) / (12*np.pi),
                0.0 
            )
        return width

    @classmethod
    def vector_nn(cls, med_mass, g=1.0):
        """Width of vector mediator decaying to neutrinos
        """
        return 3 * np.where(
            med_mass >= 2 * mass, 
            g**2 * med_mass / (24*np.pi),
            0.0
        )

    @classmethod
    def vector_dm(cls, med_mass, chi_mass=1.0, g=1.0):
        """Width of vector mediator decaying to dark matter candidates 
        """
        z = np.divide(chi_mass, med_mass)**2
        z = z.astype(np.complex128)
        return np.where(
            med_mass >= 2 * mass, 
            g**2 * med_mass * np.sqrt(1 - 4*z) * (1 + 2*z) / (12*np.pi),
            0.0
        )

    @classmethod
    def vector_total_width(cls, med_mass, chi_mass, g_q=0.25, g_chi=1.0, g_l=0.0):
        """Total width of the vector mediator
        """
        total  = cls.vector_qq(med_mass, g_q)
        total += cls.vector_ll(med_mass, g_l)
        total += cls.vector_nn(med_mass, g_l)
        total += cls.vector_dm(med_mass, chi_mass, g_chi)
        return total
    
    #Axial-Vector Mediators
    @classmethod
    def axial_qq(cls, med_mass, g = 1.0):
        width = 0
        for mass in cls._q_mass:
            z = np.divide(mass, med_mass)**2
            z = z.astype(np.complex128)
            width += np.where(
                med_mass >= 2 * mass,
                g**2 * med_mass * np.sqrt(1 - 4*z)**3 / (4*np.pi),
                0.0 
            )
        return width

    @classmethod
    def axial_ll(cls, med_mass, g = 1.0):
        width = 0
        for mass in cls._l_mass:
            z = np.divide(mass, med_mass)**2
            z = z.astype(np.complex128)
            width += np.where(
                med_mass >= 2 * mass,
                g**2 * med_mass * np.sqrt(1 - 4*z)**3 / (12*np.pi),
                0.0 
            )
        return width

    @classmethod
    def axial_nn(cls, med_mass, g = 1.0):
        return 3 * np.where(
            med_mass >= 2 * mass, 
            g**2 * med_mass / (24*np.pi),
            0.0
        )

    @classmethod
    def axial_dm(cls, med_mass, chi_mass=1.0, g = 1.0):
        z = np.divide(chi_mass, med_mass)**2
        z = z.astype(np.complex128)
        return np.where(
            med_mass >= 2 * mass, 
            g**2 * med_mass * np.sqrt(1 - 4*z)**3 / (12*np.pi),
            0.0
        )

    @classmethod
    def axial_total_width(cls, med_mass, chi_mass, g_q=0.25, g_dm=1.0, g_l=0.0):
        total  = cls.axial_qq(med_mass, g_q)
        total += cls.axial_ll(med_mass, g_l)
        total += cls.axial_nn(med_mass, g_l)
        total += cls.axial_dm(med_mass, chi_mass, g_dm)
        return total

    # Scalar Mediators
    def form_factor_s(tau):
        tau = tau.astype(np.complex128)
        return tau * (
            1 + (1 - tau)* np.arctan(np.devide(1, np.sqrt(tau - 1)))**2
        )
    def from_factor_ps(tau):
        tau = tau.astype(np.complex128)
        return tau * np.arctan(np.devide(1, np.sqrt(tau - 1)))**2

    @classmethod
    def scalar_qq(cls, med_mass, g=1.0):
        width = 0
        for mass in cls._q_mass:
            z = np.divide(mass, med_mass)**2
            z = z.astype(np.complex128)
            width += np.where(
                med_mass >= 2 * mass,
                3 * g**2 * cls._fyc(mass)**2 * med_mass * np.sqrt(1 - 4*z)**3 / (16*np.pi),
                0.0
            )
        return width
    
    @classmethod
    def scalar_gg(cls, med_mass, chi_mass=1.0, g=1.0):
        z = np.divide(cls._q_mass[5], med_mass)**2
        z = z.astype(np.complex128)
        return g**2 * med_mass * cls._as(2*chi_mass)**2 * p.where(
            med_mass >= 2 * mass, 
            cls.form_factor_s(4*z)**2 / (32*np.pi**3 * self._vev**2),
            0.0
        )

    @classmethod
    def scalar_chi(cls, med_mass, chi_mass=1.0, g=1.0):
        NotImplementedError("Will be impelemented soon")


    
    # TODO: Finish implementing all the different functions for scalar and pseudo-scalar med

       





#!/usr/bin/env python
# Copyright (c) 2021, Yacine Haddad
#
# Distributed under the 3-clause BSD license, see accompanying file LICENSE
# or https://github.com/yhaddad/lhctodd for details.

from setuptools import setup
from os import path

basedir = path.abspath(path.dirname(__file__))
with open(path.join(basedir, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()


setup(
    long_description=long_description,
    long_description_content_type='text/markdown',
)



import numpy 

class width:
    def __init__(self):
        # quark masses from PDG 2020
        self.quark_mass = [
            0.00216, # up
            0.00467, # down
            0.093,   # strange
            1.27,    # charm
            4.18,    # bottom
            172.76   # top
        ]
        # lepton masses: e mu tau from PDG 2020
        self.lepton_mass = [0.000511, 0.105658, 1.77682]
        self._vev = 246
        # fermion-Yukawa-Coupling function 
        self._fyc = lambda mf: np.sqrt(2)*mf/self.vev
        
    # Vector mediators
    def vector_qq(self, med_mass, g=1.0):
        """ Width of vector mediator decaying to quarks        
        """
        width = 0
        for mass in self.quark_mass:
            z = np.divide(mass, med_mass)**2
            width += np.where(
                med_mass >= 2 * mass,
                g**2 * med_mass * np.sqrt(1 - 4*z) * (1 + 2*z) / (4*np.pi),
                0.0 
            )
        return width

    def vector_ll(self, med_mass, g=1.0):
        """Width of vector mediator decaying to leptons
        """
        width = 0
        for mass in self.lepton_mass:
            z = np.divide(mass, med_mass)**2
            width += np.where(
                med_mass >= 2 * mass,
                g**2 * med_mass * np.sqrt(1 - 4*z) * (1 + 2*z) / (12*np.pi),
                0.0 
            )
        return width

    def vector_nn(self, med_mass, g=1.0):
        """Width of vector mediator decaying to neutrinos
        """
        return 3 * np.where(
            med_mass >= 2 * mass, 
            g**2 * med_mass / (24*np.pi),
            0.0
        )

    def vector_dm(self, med_mass, chi_mass=1.0, g=1.0):
        """Width of vector mediator decaying to dark matter candidates 
        """
        z = np.divide(chi_mass, med_mass)**2
        return np.where(
            med_mass >= 2 * mass, 
            g**2 * med_mass * np.sqrt(1 - 4*z) * (1 + 2*z) / (12*np.pi),
            0.0
        )

    def vector_total_width(self, med_mass, chi_mass, g_q=0.25, g_chi=1.0, g_l=0.0):
        """Total width of the vector mediator
        """
        total  = vector_qq(med_mass, g_q)
        total += vector_ll(med_mass, g_l)
        total += vector_nn(med_mass, g_l)
        total += vector_dm(med_mass, chi_mass, g_chi)
        return total
    
    #Axial-Vector Mediators
    def axial_qq(self, med_mass, g = 1.0):
        width = 0
        for mass in self.quark_mass:
            z = np.divide(mass, med_mass)**2
            width += np.where(
                med_mass >= 2 * mass,
                g**2 * med_mass * np.sqrt(1 - 4*z)**3 / (4*np.pi),
                0.0 
            )
        return width

    def axial_ll(self, med_mass, g = 1.0):
        width = 0
        for mass in self.lepton_mass:
            z = np.divide(mass, med_mass)**2
            width += np.where(
                med_mass >= 2 * mass,
                g**2 * med_mass * np.sqrt(1 - 4*z)**3 / (12*np.pi),
                0.0 
            )
        return width

    def axial_nn(self, med_mass, g = 1.0):
        return 3 * np.where(
            med_mass >= 2 * mass, 
            g**2 * med_mass / (24*np.pi),
            0.0
        )

    def axial_dm(self, med_mass, chi_mass=1.0, g = 1.0):
        z = np.divide(chi_mass, med_mass)**2
        return np.where(
            med_mass >= 2 * mass, 
            g**2 * med_mass * np.sqrt(1 - 4*z)**3 / (12*np.pi),
            0.0
        )

    def axial_total_width(self, med_mass, chi_mass, g_q=0.25, g_dm=1.0, g_l=0.0):
        total  = axial_qq(med_mass, g_q)
        total += axial_ll(med_mass, g_l)
        total += axial_nn(med_mass, g_l)
        total += axial_dm(med_mass, chi_mass, g_dm)
        return total

    # Scalar Mediators
    def form_factor_s(tau):
        return tau * (
            1 + (1 - tau)* np.arctan(np.devide(1, np.sqrt(tau - 1)))**2
        )
    def from_factor_ps(tau):
        return tau * np.arctan(np.devide(1, np.sqrt(tau - 1)))**2

    def scalar_qq(self, med_mass, g=1.0):
        width = 0
        for mass in self.quark_mass:
            z = np.divide(mass, med_mass)**2
            width += np.where(
                med_mass >= 2 * mass,
                3 * g**2 * self._fyk(mass)**2 * med_mass * np.sqrt(1 - 4*z)**3 / (16*np.pi),
                0.0
            )
        return width

    # TODO: Finish implementing all the different functions for scalar and pseudo-scalar med

       





from .tools import __data_path__
from .tools import dd_format
from scipy import interpolate

import lmdb
import pickle
import matplotlib.pyplot as plt


class DD:
    # Direct Detection measurements
    def __init__(self, limit_id=None, pattern=None, arxiv=None):
        self._data = None
        self._env = lmdb.open(str( __data_path__ / f"darkmatter-data"), readonly=True)
        with self._env.begin() as txn:
            if limit_id is not None:
                data = txn.get(f"{limit_id:08}".encode("ascii"))
                self._data = pickle.loads(data)
            else:
                for key, _ in txn.cursor():
                    raw = txn.get(key)
                    data = pickle.loads(raw)
                    if pattern is not None:
                        if pattern in data.__dict__.values():
                            self._data = data
                            break
                    if arxiv is not None:
                        if arxiv in data.cite:
                            self._data = data
                            break
        self._env.close()

        assert self._data is not None
        self._limit = self._data.get_limit()
        self.name = self._data.name
        self.type = self._data.type
        self.cite = "https://arxiv.org/abs/{}".format(self._data.cite)
        self._func = interpolate.interp1d(
            self._limit[:,0],
            self._limit[:,1],
            fill_value="extrapolate"
        )

    def sigma(self, mass=100):
        return self._func(mass)

    def data(self):
        return self._data.get_limit()

    def to_pandas(self):
        return pd.DataFrame(
            self._data.get_limit(),
            columns=["mass", "sigma"]
        )
    def plot(self, ax=None):
        if ax is None:
            ax = plt.gca()

        ax.plot(
            self._limit[:,0],
            self._limit[:,1],
            label= self._data.name
        )

        ax.set_xlabel("$m_{\chi}$ (GeV)")
        ax.set_ylabel(
            "{model} DM-nucleon cross-section (cm$^2$)".format(
                model=self._data.type
            )
        )
        ax.set_xscale("log")
        ax.set_yscale("log")


def plot_all(limit_type="SI", ax=None):
    _env = lmdb.open(str( __data_path__ / f"darkmatter-data"), readonly=True)
    with _env.begin() as txn:
        for key, _ in txn.cursor():
            raw = txn.get(key)
            data = pickle.loads(raw)
            if limit_type in data.type:
                limit = data.get_limit()
                if ax is None:
                    ax = plt.gca()
                ax.plot(
                    limit[:,0],
                    limit[:,1],
                    label = data.name
                )
                ax.set_xlabel("$m_{\chi}$ (GeV)")
                ax.set_ylabel(
                    "{model} DM-nucleon cross-section (cm$^2$)".format(
                        model=data.type
                    )
                )
                ax.set_xscale("log")
                ax.set_yscale("log")
    _env.close()



class sim_model:
    # simplified model
    def __init__(self, g_chi=1.0, g_quark=0.25, g_lepton=0.0, label=None):
        self.med_mass = None
        self.chi_mass = None
        self.g_chi = g_chi
        self.g_quark = g_quark
        self.g_lepton = g_lepton
        self.limit = None
        self.type = ""
        self.label = label

    def from_csv(self, filename, delimiter=","):
        print("filename : ", filename, delimiter)
        my_data = np.genfromtxt(filename, delimiter=delimiter)
        self.med_mass = my_data[:,0] # mediator
        self.chi_mass = my_data[:,1] # darkmatter
        return np.vstack([
            self.sigma(self.med_mass, self.chi_mass),
            self.chi_mass
        ]).T

    def sigma(self, med_mass, chi_mass):
        raise NotImplementedError("sigma not implemented!")

    def plot(self, ax=None):
        if ax is None:
            ax = plt.gca()

        ax.plot(
            self.chi_mass,
            self.sigma(self.med_mass, self.chi_mass),
            label= self.label
        )

        ax.set_xlabel("$m_{\chi}$")
        ax.set_ylabel(
            "{model} DM-nucleon cross-section (cm$^2$)".format(
                model=self.type
            )
        )
        ax.set_xscale("log")
        ax.set_yscale("log")


class SD(sim_model):
    """Translate LHC 2D limits on Axial-Vector mediator onto limit on DM-Nucleon cross section
    The values of the couplings should correspond to the model used to extract your limits

    Parameters
    ----------
    g_chi: coupling value to Dark Matter candidate $g_{\chi}$
    g_quark: coupling value to quarks $g_q$
    g_lepton: coupling value to leptons $g_\ell$
    label: used to label the curve when plot function is called

    Examples
    --------
    from a CSV file that contains the observed 2D limit
    $(m_{med}, m_{\chi})$, with $m_{med}$ is the mass of
    the mediator and $m_\chi$ is the one for the Dark Matter
    >>> import lhctodd as dd
    >>> model = dd.Axial(g_chi=1.0, g_qaurk=0.25)
    >>> model.from_csv("some-limit-from-lhc.csv")
    >>> model.plot()

    All in one line
    >>> dm_mass, limit_sigma = dd.Axial(g_chi=1.0, g_qaurk=0.25).from_csv("some-limit-from-lhc.csv")
    >>> plt.plot(dm_mass, limit_sigma)
    """
    def __init__(self, g_chi=1.0, g_quark=0.25, g_lepton=0.0, label=None):
        super().__init__(g_chi, g_quark, g_lepton, label)
        self.neutron_mass = 0.939

    def sigma(self, med_mass, chi_mass):
        rat = self.neutron_mass * chi_mass / (chi_mass + self.neutron_mass)
        sigma = 2.4e-42
        sigma *= np.power(self.g_quark*self.g_chi/0.25, 2)
        sigma *= np.power(1000./med_mass, 4)
        sigma *= np.power(rat,2)
        return sigma


class SI(sim_model):
    """Translate LHC 2D limits on Vector or Scalar mediators onto limit on DM-Nucleon cross section
    The values of the couplings should correspond to the model used to extract your limits

    Parameters
    ----------
    g_chi: coupling value to Dark Matter candidate $g_{\chi}$
    g_quark: coupling value to quarks $g_q$
    g_lepton: coupling value to leptons $g_\ell$
    label: used to label the curve when plot function is called

    Examples
    --------
    from a CSV file that contains the observed 2D limit
    $(m_{med}, m_{\chi})$, with $m_{med}$ is the mass of
    the mediator and $m_\chi$ is the one for the Dark Matter
    >>> import lhctodd as dd
    >>> model = dd.Vector(g_chi=1.0, g_qaurk=0.25)
    >>> model.from_csv("some-limit-from-lhc.csv")
    >>> model.plot()

    All in one line
    >>> dm_mass, limit_sigma = dd.Vector(g_chi=1.0, g_qaurk=0.25).from_csv("some-limit-from-lhc.csv")
    >>> plt.plot(dm_mass, limit_sigma)
    """

    def __init__(self, g_chi=1.0, g_quark=0.25, g_lepton=0.0, label=None):
        super().__init__(g_chi, g_quark, g_lepton, label)
        self.neutron_mass = 0.939

    def sigma(self, med_mass, chi_mass):
        rat = self.neutron_mass * chi_mass / (chi_mass + self.neutron_mass)
        sigma = 6.9e-41
        sigma *= np.power(self.g_quark*self.g_chi/0.25, 2)
        sigma *= np.power(1000./med_mass, 4)
        sigma *= np.power(rat,2)
        return sigma



"""Top-level package for lhctodd."""

__author__ = """Yacine Haddad"""
__email__ = 'yhaddad@cern.ch'
__version__ = '0.1.0'

import os
import lmdb
import pickle
from pathlib import Path
from prettytable import PrettyTable

from . import theory
from .tools import dd_format
from .tools import __data_path__
from .model import DD
from .model import SI
from .model import SD
from .model import plot_all


def list(search=None):
    env = lmdb.open(str(__data_path__ / f"darkmatter-data"), readonly=True)
    output = PrettyTable()
    output.field_names = ["id", "type", "Experiment", "arXiv"]
    with env.begin() as txn:
        for key, _ in txn.cursor():
            raw = txn.get(key)
            data = pickle.loads(raw)
            if search:
                if search in data.__dict__.values():
                    output.add_row([
                        int(key.decode()),
                        data.type,
                        data.expr,
                        data.cite
                    ])

            else:
                output.add_row([
                    int(key.decode()),
                    data.type,
                    data.expr,
                    data.cite
                ])
    env.close()
    print(output)


__all__ = ["model", "DD", "SI", "SD", "list"]
